@@ -24,7 +24,7 @@
 # Contact: Brandon Long <blong@google.com>
 #
 # This has been modified from the original software.
-# Copyright (c) 2016, 2017 Scott Kitterman <scott@kitterman.com>
+# Copyright (c) 2016, 2017, 2018 Scott Kitterman <scott@kitterman.com>
 #
 # This has been modified from the original software.
 # Copyright (c) 2017 Valimail Inc
@@ -38,6 +38,19 @@ import logging
 import re
 import time
 
+# only needed for arc
+try:
+  from authres import AuthenticationResultsHeader
+except:
+  pass
+
+# only needed for ed25519 signing/verification
+try:
+    import nacl.signing
+    import nacl.encoding
+except:
+    pass
+
 from dkim.canonicalization import (
     CanonicalizationPolicy,
     InvalidCanonicalizationPolicyError,
@@ -71,6 +84,8 @@ __all__ = [
     "MessageFormatError",
     "ParameterError",
     "ValidationError",
+    "AuthresNotFoundError",
+    "NaClNotFoundError",
     "CV_Pass",
     "CV_Fail",
     "CV_None",
@@ -141,6 +156,17 @@ class ValidationError(DKIMException):
     """Validation error."""
     pass
 
+class AuthresNotFoundError(DKIMException):
+    """ Authres Package not installed, needed for ARC """
+    pass
+
+class NaClNotFoundError(DKIMException):
+    """ Nacl package not installed, needed for ed25119 signatures """
+    pass
+
+class UnknownKeyTypeError(DKIMException):
+    """ Key type (k tag) is not known (rsa/ed25519) """
+
 def select_headers(headers, include_headers):
     """Select message header fields to be signed/verified.
 
@@ -186,6 +212,20 @@ def hash_headers(hasher, canonicalize_headers, headers, include_headers,
         hasher.update(y)
     return sign_headers
 
+def hash_headers_ed25519(pk, canonicalize_headers, headers, include_headers,
+                 sigheader, sig):
+    """Update hash for signed message header fields."""
+    hash_header = ''
+    sign_headers = select_headers(headers,include_headers)
+    # The call to _remove() assumes that the signature b= only appears
+    # once in the signature header
+    cheaders = canonicalize_headers.canonicalize_headers(
+        [(sigheader[0], RE_BTAG.sub(b'\\1',sigheader[1]))])
+    # the dkim sig is hashed with no trailing crlf, even if the
+    # canonicalization algorithm would add one.
+    for x,y in sign_headers + [(x, y.rstrip()) for x,y in cheaders]:
+        hash_header += x + y
+    return sign_headers, hash_header
 
 def validate_signature_fields(sig, mandatory_fields=[b'v', b'a', b'b', b'bh', b'd', b'h', b's'], arc=False):
     """Validate DKIM or ARC Signature fields.
@@ -249,6 +289,8 @@ def validate_signature_fields(sig, mandatory_fields=[b'v', b'a', b'b', b'bh', b'
             raise ValidationError(
               "x= value is not a decimal integer (%s)" % sig[b'x'])
         x_sign = int(sig[b'x'])
+        now = int(time.time())
+        slop = 36000 # 10H leeway for mailers with inaccurate clocks
         if x_sign < now - slop:
             raise ValidationError(
                 "x= value is past (%s)" % sig[b'x'])
@@ -343,13 +385,22 @@ def load_pk_from_dns(name, dnsfunc=get_txt):
   except InvalidTagValueList as e:
       raise KeyFormatError(e)
   try:
-      pk = parse_public_key(base64.b64decode(pub[b'p']))
-      keysize = bitsize(pk['modulus'])
+      if pub[b'k'] == b'ed25519':
+          pk = nacl.signing.VerifyKey(pub[b'p'], encoder=nacl.encoding.Base64Encoder)
+          keysize = 256
+          ktag = b'ed25519'
   except KeyError:
-      raise KeyFormatError("incomplete public key: %s" % s)
-  except (TypeError,UnparsableKeyError) as e:
-      raise KeyFormatError("could not parse public key (%s): %s" % (pub[b'p'],e))
-  return pk, keysize
+      pub[b'k'] = b'rsa'
+  if pub[b'k'] == b'rsa':
+      try:
+          pk = parse_public_key(base64.b64decode(pub[b'p']))
+          keysize = bitsize(pk['modulus'])
+      except KeyError:
+          raise KeyFormatError("incomplete public key: %s" % s)
+      except (TypeError,UnparsableKeyError) as e:
+          raise KeyFormatError("could not parse public key (%s): %s" % (pub[b'p'],e))
+      ktag = b'rsa'
+  return pk, keysize, ktag
 
 #: Abstract base class for holding messages and options during DKIM/ARC signing and verification.
 class DomainSigner(object):
@@ -500,10 +551,14 @@ class DomainSigner(object):
         h, canon_policy, headers, include_headers, header, sig)
     self.logger.debug("sign %s headers: %r" % (header_name, h.hashed()))
 
-    try:
-        sig2 = RSASSA_PKCS1_v1_5_sign(h, pk)
-    except DigestTooLargeError:
-        raise ParameterError("digest too large for modulus")
+    if self.signature_algorithm == b'rsa-sha256' or self.signature_algorithm == b'rsa-sha1':
+        try:
+            sig2 = RSASSA_PKCS1_v1_5_sign(h, pk)
+        except DigestTooLargeError:
+            raise ParameterError("digest too large for modulus")
+    elif self.signature_algorithm == b'ed25519-sha256':
+        sigobj = pk.sign(h.digest())
+        sig2 = sigobj.signature
     # Folding b= is explicity allowed, but yahoo and live.com are broken
     #header_value += base64.b64encode(bytes(sig2))
     # Instead of leaving unfolded (which lets an MTA fold it later and still
@@ -526,7 +581,7 @@ class DomainSigner(object):
   def verify_sig(self, sig, include_headers, sig_header, dnsfunc):
     name = sig[b's'] + b"._domainkey." + sig[b'd'] + b"."
     try:
-      pk, self.keysize = load_pk_from_dns(name, dnsfunc)
+      pk, self.keysize, ktag = load_pk_from_dns(name, dnsfunc)
     except KeyFormatError as e:
       self.logger.error("%s" % e)
       return False
@@ -571,16 +626,25 @@ class DomainSigner(object):
     self.signed_headers = hash_headers(
         h, canon_policy, headers, include_headers, sig_header, sig)
     self.logger.debug("signed for %s: %r" % (sig_header[0], h.hashed()))
-
-    try:
-        signature = base64.b64decode(re.sub(br"\s+", b"", sig[b'b']))
-        res = RSASSA_PKCS1_v1_5_verify(h, signature, pk)
-        self.logger.debug("%s valid: %s" % (sig_header[0], res))
-        if res and self.keysize < self.minkey:
-          raise KeyFormatError("public key too small: %d" % self.keysize)
-        return res
-    except (TypeError,DigestTooLargeError) as e:
-        raise KeyFormatError("digest too large for modulus: %s"%e)
+    signature = base64.b64decode(re.sub(br"\s+", b"", sig[b'b']))
+    if ktag == b'rsa':
+        try:
+            res = RSASSA_PKCS1_v1_5_verify(h, signature, pk)
+            self.logger.debug("%s valid: %s" % (sig_header[0], res))
+            if res and self.keysize < self.minkey:
+                raise KeyFormatError("public key too small: %d" % self.keysize)
+            return res
+        except (TypeError,DigestTooLargeError) as e:
+            raise KeyFormatError("digest too large for modulus: %s"%e)
+    elif ktag == b'ed25519':
+        try:
+            pk.verify(h.digest(), signature)
+            self.logger.debug("%s valid" % (sig_header[0]))
+            return True
+        except (nacl.exceptions.BadSignatureError) as e:
+            return False
+    else:
+        raise UnknownKeyTypeError(ktag)
 
 #: Hold messages and options during DKIM signing and verification.
 class DKIM(DomainSigner):
@@ -619,12 +683,17 @@ class DKIM(DomainSigner):
   #: @return: DKIM-Signature header field terminated by '\r\n'
   #: @raise DKIMException: when the message, include_headers, or key are badly
   #: formed.
-  def sign(self, selector, domain, privkey, identity=None,
+  def sign(self, selector, domain, privkey, signature_algorithm=None, identity=None,
         canonicalize=(b'relaxed',b'simple'), include_headers=None, length=False):
-    try:
-        pk = parse_pem_private_key(privkey)
-    except UnparsableKeyError as e:
-        raise KeyFormatError(str(e))
+    if signature_algorithm:
+        self.signature_algorithm = signature_algorithm
+    if self.signature_algorithm == b'rsa-sha256' or self.signature_algorithm == b'rsa-sha1':
+        try:
+            pk = parse_pem_private_key(privkey)
+        except UnparsableKeyError as e:
+            raise KeyFormatError(str(e))
+    elif self.signature_algorithm == b'ed25519-sha256':
+        pk = nacl.signing.SigningKey(privkey, encoder=nacl.encoding.Base64Encoder)
 
     if identity is not None and not identity.endswith(domain):
         raise ParameterError("identity must end with domain")
@@ -765,27 +834,59 @@ class ARC(DomainSigner):
   #: @param selector: the DKIM selector value for the signature
   #: @param domain: the DKIM domain value for the signature
   #: @param privkey: a PKCS#1 private key in base64-encoded text form
-  #: @param auth_results: RFC 7601 Authentication-Results header value for the message
-  #: @param chain_validation_status: CV_Pass, CV_Fail, CV_None
+  #: @param srv_id: an srv_id for identitfying AR headers to sign & extract cv from
   #: @param include_headers: a list of strings indicating which headers
   #: are to be signed (default rfc4871 recommended headers)
   #: @return: list of ARC set header fields
   #: @raise DKIMException: when the message, include_headers, or key are badly
   #: formed.
-  def sign(self, selector, domain, privkey, auth_results, chain_validation_status,
-           include_headers=None, timestamp=None, standardize=False):
+  def sign(self, selector, domain, privkey, srv_id, include_headers=None,
+           timestamp=None, standardize=False):
+
+    # check if authres has been imported
+    try:
+        AuthenticationResultsHeader
+    except:
+        self.logger.debug("authres package not installed")
+        raise AuthresNotFoundError
+
     try:
         pk = parse_pem_private_key(privkey)
     except UnparsableKeyError as e:
         raise KeyFormatError(str(e))
 
+    # extract, parse, filter & group AR headers
+    ar_headers = [res.strip() for [ar, res] in self.headers if ar == b'Authentication-Results']
+    grouped_headers = [(res, AuthenticationResultsHeader.parse('Authentication-Results: ' + res.decode('utf-8')))
+                       for res in ar_headers]
+    auth_headers = [res for res in grouped_headers if res[1].authserv_id == srv_id.decode('utf-8')]
+
+    if len(auth_headers) == 0:
+      self.logger.debug("no AR headers found, chain terminated")
+      return b''
+
+    # consolidate headers
+    results_lists = [raw.replace(srv_id + b';', b'').strip() for (raw, parsed) in auth_headers]
+    results_lists = [tags.split(b';') for tags in results_lists]
+    results = [tag.strip() for sublist in results_lists for tag in sublist]
+    auth_results = srv_id + b'; ' + b';\r\n  '.join(results)
+
+    # extract cv
+    parsed_auth_results = AuthenticationResultsHeader.parse('Authentication-Results: ' + auth_results.decode('utf-8'))
+    arc_results = [res for res in parsed_auth_results.results if res.method == 'arc']
+    if len(arc_results) == 0:
+      self.logger.debug("no AR arc stamps found, chain terminated")
+      return b''
+    elif len(arc_results) != 1:
+      self.logger.debug("multiple AR arc stamps found, failing chain")
+      chain_validation_status = CV_Fail
+    else:
+      chain_validation_status = arc_results[0].result.lower().encode('utf-8')
+
     # Setup headers
     if include_headers is None:
         include_headers = self.default_sign_headers()
 
-    if b'arc-authentication-results' not in include_headers:
-        include_headers.append(b'arc-authentication-results')
-
     include_headers = tuple([x.lower() for x in include_headers])
 
     # record what verify should extract
@@ -811,7 +912,10 @@ class ARC(DomainSigner):
       raise ParameterError("cv=none not allowed on instance %d" % instance)
 
     new_arc_set = []
-    arc_headers = [y for x,y in arc_headers_w_instance]
+    if chain_validation_status != CV_Fail:
+      arc_headers = [y for x,y in arc_headers_w_instance]
+    else: # don't include previous sets for a failed/invalid chain
+      arc_headers = []
 
     # Compute ARC-Authentication-Results
     aar_value = ("i=%d; " % instance).encode('utf-8') + auth_results
@@ -869,6 +973,11 @@ class ARC(DomainSigner):
     as_include_headers = [x[0].lower() for x in arc_headers]
     as_include_headers.reverse()
 
+    # if our chain is failing or invalid, we only grab the most recent set
+    # reversing the order of the headers accomplishes this
+    if chain_validation_status == CV_Fail:
+      self.headers.reverse()
+
     res = self.gen_header(as_fields, as_include_headers, canon_policy,
                            b"ARC-Seal", pk, standardize)
 
@@ -887,7 +996,7 @@ class ARC(DomainSigner):
   #: @param dnsfunc: an optional function to lookup TXT resource records
   #: for a DNS domain.  The default uses dnspython or pydns.
   #: @return: True if signature verifies or False otherwise
-  #: @return: three-tuple of (CV Result (CV_Pass, CV_Fail or CV_None), list of
+  #: @return: three-tuple of (CV Result (CV_Pass, CV_Fail, CV_None or None, for a chain that has ended), list of
   #: result dictionaries, result reason)
   #: @raise DKIMException: when the message, signature, or key are badly formed
   def verify(self,dnsfunc=get_txt):
@@ -907,10 +1016,10 @@ class ARC(DomainSigner):
     if not result_data[0]['ams-valid']:
         return CV_Fail, result_data, "Most recent ARC-Message-Signature did not validate"
     for result in result_data:
-      if not result['as-valid']:
-        return CV_Fail, result_data, "ARC-Seal[%d] did not validate" % result['instance']
       if result['cv'] == CV_Fail:
-        return CV_Fail, result_data, "ARC-Seal[%d] reported failure" % result['instance']
+        return None, result_data, "ARC-Seal[%d] reported failure, the chain is terminated" % result['instance']
+      elif not result['as-valid']:
+        return CV_Fail, result_data, "ARC-Seal[%d] did not validate" % result['instance']
       elif (result['instance'] == 1) and (result['cv'] != CV_None):
         return CV_Fail, result_data, "ARC-Seal[%d] reported invalid status %s" % (result['instance'], result['cv'])
       elif (result['instance'] != 1) and (result['cv'] == CV_None):
@@ -968,7 +1077,7 @@ class ARC(DomainSigner):
 
     self.logger.debug("ams sig[%d]: %r" % (instance, sig))
 
-    validate_signature_fields(sig, [b'i', b'a', b'b', b'c', b'bh', b'd', b'h', b's'], True)
+    validate_signature_fields(sig, [b'i', b'a', b'b', b'bh', b'd', b'h', b's'], True)
     output['ams-domain'] = sig[b'd']
     output['ams-selector'] = sig[b's']
 
@@ -977,7 +1086,18 @@ class ARC(DomainSigner):
         raise ParameterError("The Arc-Message-Signature MUST NOT sign ARC-Seal")
 
     ams_header = (b'ARC-Message-Signature', b' ' + ams_value)
-    ams_valid = self.verify_sig(sig, include_headers, ams_header, dnsfunc)
+
+
+    # we can't use the AMS provided above, as it's already been canonicalized relaxed
+    # for use in validating the AS.  However the AMS is included in the AMS itself,
+    # and this can use simple canonicalization
+    raw_ams_header = [(x, y) for (x, y) in self.headers if x.lower() == b'arc-message-signature'][0]
+
+    try:
+      ams_valid = self.verify_sig(sig, include_headers, raw_ams_header, dnsfunc)
+    except DKIMException as e:
+      self.logger.error("%s" % e)
+      ams_valid = False
 
     output['ams-valid'] = ams_valid
     self.logger.debug("ams valid: %r" % ams_valid)
@@ -990,7 +1110,7 @@ class ARC(DomainSigner):
 
     self.logger.debug("as sig[%d]: %r" % (instance, sig))
 
-    validate_signature_fields(sig, [b'i', b'a', b'b', b'cv', b'd', b's', b't'], True)
+    validate_signature_fields(sig, [b'i', b'a', b'b', b'cv', b'd', b's'], True)
     output['as-domain'] = sig[b'd']
     output['as-selector'] = sig[b's']
     output['cv'] = sig[b'cv']
@@ -998,7 +1118,11 @@ class ARC(DomainSigner):
     as_include_headers = [x[0].lower() for x in arc_headers]
     as_include_headers.reverse()
     as_header = (b'ARC-Seal', b' ' + as_value)
-    as_valid = self.verify_sig(sig, as_include_headers[:-1], as_header, dnsfunc)
+    try:
+      as_valid = self.verify_sig(sig, as_include_headers[:-1], as_header, dnsfunc)
+    except DKIMException as e:
+      self.logger.error("%s" % e)
+      as_valid = False
 
     output['as-valid'] = as_valid
     self.logger.debug("as valid: %r" % as_valid)
@@ -1045,8 +1169,7 @@ dkim_sign = sign
 dkim_verify = verify
 
 def arc_sign(message, selector, domain, privkey,
-             auth_results, chain_validation_status,
-             signature_algorithm=b'rsa-sha256',
+             srv_id, signature_algorithm=b'rsa-sha256',
              include_headers=None, timestamp=None,
              logger=None, standardize=False):
     """Sign an RFC822 message and return the ARC set header lines for the next instance
@@ -1054,19 +1177,19 @@ def arc_sign(message, selector, domain, privkey,
     @param selector: the DKIM selector value for the signature
     @param domain: the DKIM domain value for the signature
     @param privkey: a PKCS#1 private key in base64-encoded text form
-    @param auth_results: the RFC 7601 authentication-results header field value for this instance
-    @param chain_validation_status: the validation status of the existing chain on the message (P (pass), F (fail)) or N (none) for no existing chain
+    @param srv_id: the authserv_id used to identify the ADMD's AR headers
     @param signature_algorithm: the signing algorithm to use when signing
     @param include_headers: a list of strings indicating which headers are to be signed (default all headers not listed as SHOULD NOT sign)
     @param logger: a logger to which debug info will be written (default None)
     @return: A list containing the ARC set of header fields for the next instance
     @raise DKIMException: when the message, include_headers, or key are badly formed.
     """
+
     a = ARC(message,logger=logger,signature_algorithm=signature_algorithm)
     if not include_headers:
         include_headers = a.default_sign_headers()
-    return a.sign(selector, domain, privkey, auth_results, chain_validation_status,
-                  include_headers=include_headers, timestamp=timestamp, standardize=standardize)
+    return a.sign(selector, domain, privkey, srv_id, include_headers=include_headers,
+                  timestamp=timestamp, standardize=standardize)
 
 def arc_verify(message, logger=None, dnsfunc=get_txt, minkey=1024):
     """Verify the ARC chain on an RFC822 formatted message.
